// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file SoapHandler.qm SOAP handler module providing the SoapHandler class to be registered as a handler with the Qore HttpServer module

/*  SoapHandler.qm Copyright (C) 2012 - 2014 David Nichols

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/


// make sure we have the required qore version

// requires XML functionality

// requires the WSDL module

// need mime definitions

// need definitions from the HttpServer


/*  Version History
    * 2014-03-04 v0.2.5: David Nichols <david@qore.org>: added support for matching
                         requests with soap action values

    * 2014-01-14 v0.2.5: David Nichols <david@qore.org>: added the err_func argument 
    *                    so SOAP response serialization errors can be logged at the source

    * 2012-05-31 v0.2.4: David Nichols <david@qore.org>: updated to a user module

    *            v0.2.3: return correct soap fault structure according to request

    *            v0.2.2: soap 1.2 improvements

    *            v0.2.1: matched to new WSDL and HttpServer implementation

    *            v0.2.0: better WSDL support

    *            v0.1.0: initial WSDL-based SOAP support (still incomplete)
*/

/** @mainpage SoapHandler Module
    
    @section soaphandlerintro Introduction to the SoapHandler Module

    This module implements server-side support for the <a href="http://en.wikipedia.org/wiki/SOAP">SOAP protocol</a> for serialization/deserialization of message data.

    This module provides the @ref SoapHandler::SoapHandler "SoapHandler" class which can be used to provide an RPC handler for the HttpServer class provided by the HttpServer module.
*/

//! main SoapHandler namespace
namespace SoapHandler {
    //! SoapHandler implementation; to be registered as a request handler in the HttpServer class
class SoapHandler : public AbstractHttpRequestHandler {

public:
        //! version of the SoapHandler implementation
        const Version = "0.2.5";

        //! @cond nodoc
        private :
            hash methods;
            mi;

            // hash of WebService objects (values) per path (keys)
            hash ws;

            // path -> top-level-method-name -> method
            hash pathmethods;

            // soapaction map: action -> operation
            hash sam;

            // path -> soapaction
            hash psam;

            int loglevel;

            // if True then verbose exception info will be logged
            bool debug;

            // a closure/call reference to get the log message and/or process arguments in incoming requests
            __7_ code getLogMessage;

            // for atomicity when adding / removing methods
            RWLock rwl();

            // hash of methods by path
            hash th;

public:
        //! @endcond

        //! creates the handler with the given method list
/** @param auth an authentication object (use new AbstractAuthenticator() for no authentication)
            @param getLogMessage an optional closure or call reference to be called when an incoming request is received; if this is set then it will be called with the following arguments: a context hash (see HttpServer::AbstractHttpRequestHandler::handleRequest() for a description of the context hash), the method definition as passed in the methods argument to this constructor, and a reference to the arguments in this call
            @param dbg this parameter is set to @ref Qore::True "True", then additional information will be logged when errors occur
         */
        constructor(AbstractAuthenticator auth, __7_ code getLogMessage, bool dbg = False);


        //! adds a method to the handler dynamically
/** @param ws the WebService object for the method
            @param op the WSOperation object for the web service operation the method corresponds to
            @param func a call reference, a closure, or a string function name to call with the deserialized arguments to the method; the return value will be serialized to SOAP according to the WSDL and sent back to the caller
            @param help optional help text for the method
            @param logopt log options which can be used by a custom logger (see the getLogMessage parameter in the constructor)
            @param cmark an optional "context marker" for the method
            @param path an optional path for the method
            @param err_func a call reference, a closure, or a string function name to call with error information if an exception is thrown with SOAP data
         */
        addMethod(WebService ws, WSOperation op, any func, __7_ string help, __7_ int logopt, any cmark, __7_ string path, any err_func);


        //! call to remove all service mappings for services with the given path
        removeServicesWithPath(string path);


        //! @cond nodoc
        // don't reimplement this method; fix/enhance it in the module
        final private addMethodInternal(WebService ws, hash method);


        private __7_ hash help(hash cx);


        // don't reimplement this method; fix/enhance it in the module
        final private log(hash cx, string str);


        // don't reimplement this method; fix/enhance it in the module

private:
          static string makeSoapFaultResponse(bool soap12, string err, string desc);
public:


        // don't reimplement this method; fix/enhance it in the module
        final private hash callOperation(hash cx, any args, __7_ string soapaction);


        // method called by HttpServer
        // don't reimplement this method; fix/enhance it in the module
        final hash handleRequest(hash cx, hash hdr, __7_ data body);

        //! @endcond
    };
};
